@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2012-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2012-2023. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -33,6 +33,7 @@
 -export([cmp_term/2]).
 -export([map_to_tuple_keys/1, term_type/1, map_hashmap_children/1,
          map_next/3]).
+-export([mc_iterator/1, mc_refill/1]).
 -export([open_port/2, port_command/3, port_connect/2, port_close/1,
 	 port_control/3, port_call/3, port_info/1, port_info/2]).
 
@@ -47,17 +48,19 @@
 -export([is_process_executing_dirty/1]).
 -export([dirty_process_handle_signals/1]).
 
--export([release_literal_area_switch/0, wait_release_literal_area_switch/1]).
+-export([wait_release_literal_area_switch/1]).
 
 -export([purge_module/2]).
 
--export([flush_monitor_messages/3]).
+-export([flush_monitor_messages/3,
+         '@flush_monitor_messages_refopt'/0]).
 
 -export([await_result/1, gather_io_bytes/2]).
 
 -export([time_unit/0, perf_counter_unit/0]).
 
--export([is_system_process/1]).
+-export([is_system_process/1,
+         set_code_and_literal_cleaner_prio/1]).
 
 -export([await_microstate_accounting_modifications/3,
 	 gather_microstate_accounting_result/2]).
@@ -67,8 +70,9 @@
 -export([dist_ctrl_put_data/2]).
 
 -export([get_dflags/0]).
+-export([get_creation/0]).
 -export([new_connection/1]).
--export([abort_connection/2]).
+-export([abort_pending_connection/2]).
 
 -export([scheduler_wall_time/1, system_flag_scheduler_wall_time/1,
          gather_sched_wall_time_result/1,
@@ -89,7 +93,7 @@
 
 -export([process_flag/3]).
 
--export([create_dist_channel/4]).
+-export([create_dist_channel/3]).
 
 -export([erase_persistent_terms/0]).
 
@@ -103,6 +107,22 @@
 -export([ets_lookup_binary_info/2, ets_super_user/1, ets_info_binary/1,
          ets_raw_first/1, ets_raw_next/2]).
 
+-export([get_internal_state_blocked/1]).
+
+-export([spawn_request/4, spawn_init/1, dist_spawn_request/4, dist_spawn_init/1]).
+
+-export([crasher/6]).
+
+-export([prepare_loading/2, beamfile_chunk/2, beamfile_module_md5/1]).
+
+-export([no_aux_work_threads/0]).
+
+-export([binary_to_integer/2, list_to_integer/2]).
+
+-export([dynamic_node_name/0, dynamic_node_name/1]).
+
+-export([term_to_string/1, term_to_string/2]).
+
 %%
 %% Await result of send to port
 %%
@@ -245,7 +265,7 @@ port_info(_Result, _Item) ->
     erlang:nif_error(undefined).
 
 -spec request_system_task(Pid, Prio, Request) -> 'ok' when
-      Prio :: 'max' | 'high' | 'normal' | 'low',
+      Prio :: 'max' | 'high' | 'normal' | 'low' | 'inherit',
       Type :: 'major' | 'minor',
       Request :: {'garbage_collect', term(), Type}
 	       | {'check_process_code', term(), module()}
@@ -287,10 +307,8 @@ check_process_code(Pid, Module, OptionList)  ->
     Async = get_cpc_opts(OptionList, sync),
     case Async of
 	{async, ReqId} ->
-	    {priority, Prio} = erlang:process_info(erlang:self(),
-						   priority),
 	    erts_internal:request_system_task(Pid,
-					      Prio,
+					      inherit,
 					      {check_process_code,
 					       ReqId,
 					       Module}),
@@ -300,11 +318,9 @@ check_process_code(Pid, Module, OptionList)  ->
 		true ->
 		    erts_internal:check_process_code(Module);
 		false ->
-		    {priority, Prio} = erlang:process_info(erlang:self(),
-							   priority),
 		    ReqId = erlang:make_ref(),
 		    erts_internal:request_system_task(Pid,
-						      Prio,
+						      inherit,
 						      {check_process_code,
 						       ReqId,
 						       Module}),
@@ -315,14 +331,15 @@ check_process_code(Pid, Module, OptionList)  ->
 	    end
     end.
 
-% gets async opt and verify valid option list
+%% gets async opt and verify valid option list
 get_cpc_opts([{async, _ReqId} = AsyncTuple | Options], _OldAsync) ->
     get_cpc_opts(Options, AsyncTuple);
-get_cpc_opts([{allow_gc, AllowGC} | Options], Async) when AllowGC == true;
-							  AllowGC == false ->
+get_cpc_opts([{allow_gc, AllowGC} | Options], Async) when is_boolean(AllowGC) ->
     get_cpc_opts(Options, Async);
 get_cpc_opts([], Async) ->
-    Async.
+    Async;
+get_cpc_opts(_, _) ->
+    error(bad_option).
 
 -spec check_dirty_process_code(Pid, Module) -> Result when
       Result :: boolean() | 'normal' | 'busy',
@@ -343,11 +360,6 @@ is_process_executing_dirty(_Pid) ->
 dirty_process_handle_signals(_Pid) ->
     erlang:nif_error(undefined).
 
--spec release_literal_area_switch() -> 'true' | 'false'.
-
-release_literal_area_switch() ->
-    erlang:nif_error(undefined).
-
 -spec wait_release_literal_area_switch(WaitMsg) -> 'true' | 'false' when
       WaitMsg :: term().
 
@@ -355,7 +367,7 @@ wait_release_literal_area_switch(WaitMsg) ->
     %% release_literal_area_switch() traps to here
     %% when it needs to wait
     receive WaitMsg -> ok end,
-    erts_internal:release_literal_area_switch().
+    erts_literal_area_collector:release_area_switch().
 
 -spec purge_module(Module, Op) -> boolean() when
       Module :: module(),
@@ -423,7 +435,7 @@ map_hashmap_children(_M) ->
 
 %% return the next assoc in the iterator and a new iterator
 -spec map_next(I, M, A) -> {K,V,NI} | list() when
-      I :: non_neg_integer(),
+      I :: non_neg_integer() | list(),
       M :: map(),
       K :: term(),
       V :: term(),
@@ -433,6 +445,61 @@ map_hashmap_children(_M) ->
 map_next(_I, _M, _A) ->
     erlang:nif_error(undefined).
 
+%% Introduced in Erlang/OTP 26. This function is a helper, called from
+%% code generated by the compiler. It must be kept compatible as long
+%% code calling this helper can still be loaded.
+-spec mc_iterator(MapOrIter) -> NI when
+      MapOrIter :: map() | maps:iterator(),
+      NI :: term().
+
+mc_iterator(Map) when is_map(Map) ->
+    erts_internal:map_next(0, Map, iterator);
+mc_iterator([Path | Map]) ->
+    %% This is probably an iterator.
+    try erts_internal:map_next(Path, Map, iterator) of
+        Iter ->
+            Iter
+    catch
+        error:badarg ->
+            []
+    end;
+mc_iterator(MapIter) ->
+    %% Possible "used" iterator. Must validate it.
+    case is_map_iter(MapIter) of
+        true -> MapIter;
+        false -> []
+    end.
+
+is_map_iter({_, _, Iter}) ->
+    is_map_iter(Iter);
+is_map_iter(Iter) ->
+    case Iter of
+        [Path | Map] ->
+            try erts_internal:map_next(Path, Map, iterator) of
+                _ ->
+                    true
+            catch
+                error:badarg ->
+                    false
+            end;
+        none -> true;
+        _ -> false
+    end.
+
+%% Introduced in Erlang/OTP 26. This function is a helper, called from
+%% code generated by the compiler. It must be kept compatible as long
+%% code calling this helper can still be loaded.
+-spec mc_refill(IM) -> {K,V,NI} when
+      IM :: nonempty_improper_list(I, M),
+      I :: non_neg_integer(),
+      M :: map(),
+      K :: term(),
+      V :: term(),
+      NI :: term().
+
+mc_refill([Path | Map]) ->
+    erts_internal:map_next(Path, Map, iterator).
+
 -spec erts_internal:flush_monitor_messages(Ref, Multi, Res) -> term() when
       Ref :: reference(),
       Multi :: boolean(),
@@ -454,6 +521,18 @@ flush_monitor_messages(Ref, Multi, Res) when is_reference(Ref) ->
 	    Res
     end.
 
+-spec '@flush_monitor_messages_refopt'() -> ok.
+'@flush_monitor_messages_refopt'() ->
+    %% Enables reference optimization in flush_monitor_messages/3. Note that we
+    %% both body- and tail-call it to ensure that the reference isn't cleared,
+    %% in case the caller to demonitor/2 wants to continue using it.
+    %%
+    %% This never actually runs and is only used to trigger the optimization,
+    %% see the module comment in beam_ssa_recv for details.
+    Ref = make_ref(),
+    flush_monitor_messages(Ref, true, ok),
+    flush_monitor_messages(Ref, true, ok).
+
 -spec erts_internal:time_unit() -> pos_integer().
 
 time_unit() ->
@@ -470,6 +549,26 @@ perf_counter_unit() ->
 is_system_process(_Pid) ->
     erlang:nif_error(undefined).
 
+set_code_and_literal_cleaner_prio(Prio) ->
+    Ref1 = make_ref(),
+    erts_code_purger ! {change_prio, self(), Ref1, Prio},
+    Ref2 = make_ref(),
+    LAC = find_lac(),
+    LAC ! {change_prio, self(), Ref2, Prio},
+    [{code_purger, receive {Ref1, OP1} -> OP1 end},
+     {literal_area_collector, receive {Ref2, OP2} -> OP2 end}].
+
+find_lac() ->
+    find_lac(erlang:processes()).
+
+find_lac([Pid|Pids]) ->
+    case process_info(Pid, initial_call) of
+        {initial_call, {erts_literal_area_collector, start, 0}} ->
+            Pid;
+        _ ->
+            find_lac(Pids)
+    end.
+
 -spec await_microstate_accounting_modifications(Ref, Result, Threads) -> boolean() when
       Ref :: reference(),
       Result :: boolean(),
@@ -501,7 +600,7 @@ microstate_accounting(Ref, Threads) ->
                    | existing | existing_processes | existing_ports
                    | new | new_processes | new_ports,
       How :: boolean(),
-      FlagList :: [].
+      FlagList :: list().
 trace(_PidSpec, _How, _FlagList) ->
     erlang:nif_error(undefined).
 
@@ -517,7 +616,7 @@ trace(_PidSpec, _How, _FlagList) ->
                  | boolean()
                  | restart
                  | pause,
-      FlagList :: [ ].
+      FlagList :: list().
 trace_pattern(_MFA, _MatchSpec, _FlagList) ->
     erlang:nif_error(undefined).
 
@@ -562,16 +661,20 @@ dist_ctrl_put_data(DHandle, IoList) ->
 get_dflags() ->
     erlang:nif_error(undefined).
 
+-spec erts_internal:get_creation() -> pos_integer().
+get_creation() ->
+    erlang:nif_error(undefined).
+
 -spec erts_internal:new_connection(Node) -> ConnId when
       Node :: atom(),
       ConnId :: {integer(), erlang:dist_handle()}.
 new_connection(_Node) ->
     erlang:nif_error(undefined).
 
--spec erts_internal:abort_connection(Node, ConnId) -> boolean() when
+-spec erts_internal:abort_pending_connection(Node, ConnId) -> boolean() when
       Node :: atom(),
       ConnId :: {integer(), erlang:dist_handle()}.
-abort_connection(_Node, _ConnId) ->
+abort_pending_connection(_Node, _ConnId) ->
     erlang:nif_error(undefined).
 
 %% Scheduler wall time
@@ -701,17 +804,17 @@ process_display(_Pid, _Type) ->
 process_flag(_Pid, _Flag, _Value) ->
     erlang:nif_error(undefined).
 
--spec create_dist_channel(Node, DistCtrlr, Flags, Ver) -> Result when
+-spec create_dist_channel(Node, DistCtrlr, {Flags, Cr}) -> Result when
       Node :: atom(),
       DistCtrlr :: port() | pid(),
       Flags :: integer(),
-      Ver :: integer(),
+      Cr :: pos_integer(),
       Result :: {'ok', erlang:dist_handle()}
               | {'message', reference()}
               | 'badarg'
               | 'system_limit'.
                                  
-create_dist_channel(_Node, _DistCtrlr, _Flags, _Ver) ->
+create_dist_channel(_Node, _DistCtrlr, _Tpl) ->
     erlang:nif_error(undefined).
 
 -spec erase_persistent_terms() -> 'ok'.
@@ -755,7 +858,7 @@ spawn_system_process(_Mod, _Func, _Args) ->
 %%
 
 -spec ets_lookup_binary_info(Tab, Key) -> BinInfo when
-      Tab :: ets:tab(),
+      Tab :: ets:table(),
       Key :: term(),
       BinInfo :: [{non_neg_integer(), non_neg_integer(), non_neg_integer()}].
 
@@ -769,20 +872,20 @@ ets_super_user(_Bool) ->
     erlang:nif_error(undef).
 
 -spec ets_raw_first(Tab) -> term() when
-      Tab :: ets:tab().
+      Tab :: ets:table().
 
 ets_raw_first(_Tab) ->
     erlang:nif_error(undef).
     
 -spec ets_raw_next(Tab, Key) -> term() when
-      Tab :: ets:tab(),
+      Tab :: ets:table(),
       Key :: term().
 
 ets_raw_next(_Tab, _Key) ->
     erlang:nif_error(undef).
 
 -spec ets_info_binary(Tab) -> BinInfo when
-      Tab :: ets:tab(),
+      Tab :: ets:table(),
       BinInfo :: [{non_neg_integer(), non_neg_integer(), non_neg_integer()}].
 
 ets_info_binary(Tab) ->
@@ -816,3 +919,177 @@ ets_info_binary_iter(Tab, Key, Acc) ->
                  [_|_] = BIL -> BIL ++ Acc
              end,
     ets_info_binary_iter(Tab, erts_internal:ets_raw_next(Tab, Key), NewAcc).
+
+-spec get_internal_state_blocked(Arg :: term()) -> term().
+
+get_internal_state_blocked(Arg) ->
+    erlang:system_flag(multi_scheduling, block),
+    Result = try
+                 erts_debug:get_internal_state({Arg,
+                                                blocked})
+             after
+                 erlang:system_flag(multi_scheduling, unblock)
+             end,
+    Result.
+
+-spec spawn_request(Module, Function, Args, Opts) -> Res when
+      Module :: module(),
+      Function :: atom(),
+      Args :: [term()],
+      Opts :: [term()],
+      Res :: reference() | 'badarg' | 'badopt'.
+
+spawn_request(_Module, _Function, _Args, _Opts) ->
+    erlang:nif_error(undef).
+
+-spec spawn_init({Module, Function, Args}) -> Res when
+      Module :: module(),
+      Function :: atom(),
+      Args :: [term()],
+      Res :: term().
+
+spawn_init({M, F, A}) ->
+    apply(M, F, A).
+
+-spec dist_spawn_request(Node, MFA, Opts, spawn_request) -> Res when
+      Node :: node(),
+      MFA :: {Module, Function, Args},
+      Module :: module(),
+      Function :: atom(),
+      Args :: [term()],
+      Opts :: [term()],
+      Res :: reference() | 'badarg';
+                        (Node, MFA, Opts, spawn_opt) -> Res when
+      Node :: node(),
+      MFA :: {Module, Function, Args},
+      Module :: module(),
+      Function :: atom(),
+      Args :: [term()],
+      Opts :: [term()],
+      Res :: {reference(), boolean()} | 'badarg'.
+
+dist_spawn_request(_Node, _MFA, _Opts, _Type) ->
+    erlang:nif_error(undef).
+
+-spec dist_spawn_init(MFA) -> Res when
+      MFA :: {Module, Function, non_neg_integer()},
+      Module :: module(),
+      Function :: atom(),
+      Res :: term().
+
+dist_spawn_init(MFA) ->
+    %%
+    %% The argument list is passed as a message
+    %% to the newly created process. This since
+    %% it might be large and require a substantial
+    %% amount of work to decode. This way we put
+    %% this work on the newly created process
+    %% (which can execute in parallel with all
+    %% other tasks) instead of on the distribution
+    %% channel code which is a bottleneck in the
+    %% system.
+    %% 
+    %% erl_create_process() ensures that the
+    %% argument list to use in apply is
+    %% guaranteed to be the first message in the
+    %% message queue.
+    %%
+    {M, F, _NoA} = MFA,
+    receive
+        A ->
+            erlang:apply(M, F, A)
+    end.
+
+%%
+%% Failed distributed spawn(), spawn_link(), spawn_monitor(), spawn_opt()
+%% spawns a dummy process executing the crasher/6 function...
+%%
+
+crasher(Node,Mod,Fun,Args,[],Reason) ->
+    error_logger:warning_msg("** Can not start ~w:~w,~w on ~w **~n",
+			     [Mod,Fun,Args,Node]),
+    erlang:exit(Reason);
+crasher(Node,Mod,Fun,Args,Opts,Reason) ->
+    error_logger:warning_msg("** Can not start ~w:~w,~w (~w) on ~w **~n",
+			     [Mod,Fun,Args,Opts,Node]),
+    erlang:exit(Reason).
+
+%%
+%% Actual BIF for erlang:prepare_loading/2, which decompresses the module when
+%% necessary to save us from having to do it in C code.
+%%
+-spec prepare_loading(Module, Code) -> PreparedCode | {error, Reason} when
+      Module :: module(),
+      Code :: binary(),
+      PreparedCode :: erlang:prepared_code(),
+      Reason :: badfile.
+prepare_loading(_Module, _Code) ->
+    erlang:nif_error(undefined).
+
+-spec beamfile_chunk(Bin, Chunk) -> binary() | undefined when
+      Bin :: binary(),
+      Chunk :: string().
+beamfile_chunk(_Bin, _Chunk) ->
+    erlang:nif_error(undefined).
+
+-spec beamfile_module_md5(binary()) -> binary() | undefined.
+beamfile_module_md5(_Bin) ->
+    erlang:nif_error(undefined).
+
+-spec no_aux_work_threads() -> pos_integer().
+
+no_aux_work_threads() ->
+    erlang:nif_error(undefined).
+
+%% Helper BIF for binary_to_integer/{1,2}.
+
+-spec binary_to_integer(Bin, Base) -> integer() | big | 'badarg' when
+      Bin :: binary(),
+      Base :: 2..36.
+binary_to_integer(_Bin, _Base) ->
+    erlang:nif_error(undefined).
+
+%% Helper BIF for list_to_integer/{1,2}.
+
+-spec list_to_integer(List, Base) ->
+          {integer(),list()} | 'big' | 'badarg' | 'no_integer' | 'not_a_list' when
+      List :: [any()],
+      Base :: 2..36.
+list_to_integer(_List, _Base) ->
+    erlang:nif_error(undefined).
+
+%%
+%% Is dynamic node name enabled?
+%%
+-spec dynamic_node_name() -> boolean().
+
+dynamic_node_name() ->
+    case persistent_term:get({?MODULE, dynamic_node_name}, false) of
+        false -> false;
+        _ -> true
+    end.
+
+%%
+%% Save whether dynamic node name is enabled or not.
+%%
+-spec dynamic_node_name(boolean()) -> ok.
+
+dynamic_node_name(true) ->
+    persistent_term:put({?MODULE, dynamic_node_name}, true);
+dynamic_node_name(false) ->
+    case dynamic_node_name() of
+        false -> ok;
+        _ -> _ = persistent_term:erase({?MODULE, dynamic_node_name}), ok
+    end.
+
+-spec term_to_string(T :: term()) -> string().
+
+term_to_string(T) ->
+    term_to_string(T, undefined).
+
+-spec term_to_string(T, Limit) -> string() when
+    T :: term(),
+    Limit :: undefined | pos_integer().
+
+term_to_string(_T, _Limit) ->
+    erlang:nif_error(undefined).
